#ifndef MSP_GL_VERTEXBUILDER_H_
#define MSP_GL_VERTEXBUILDER_H_

#include <map>
#include "color.h"
#include "matrix.h"
#include "vector.h"
#include "vertexformat.h"

namespace Msp {
namespace GL {

/**
Base class for classes that build vertices from a series of function calls.
The operating model closely follows that of OpenGL immediate mode: vertex
attributes can be specified at any time, and when a vertex() function is
called, a vertex is created with the active attribute values.

A derived class must overload the 4-argument vertex_() function to process the
data.  Attributes can be read from protected member variables.
*/
class VertexBuilder
{
public:
	class PushMatrix
	{
	private:
		VertexBuilder &bld;
		Matrix mtx;

	public:
		PushMatrix(VertexBuilder &b): bld(b), mtx(bld.mtx) { }
		~PushMatrix() { bld.mtx = mtx; }
	};

protected:
	Matrix mtx;
	std::vector<Vector4> attr;

public:
	VertexBuilder() { normal(0, 0, 1); }

	virtual ~VertexBuilder() { }

	void set_matrix(const Matrix &m)
	{ mtx = m; }

	void transform(const Matrix &m)
	{ mtx *= m; }

	const Matrix &get_matrix() const
	{ return mtx; }

	void vertex(float x, float y)
	{ vertex(x, y, 0, 1); }

	void vertex(float x, float y, float z)
	{ vertex(x, y, z, 1); }

	void vertex(float x, float y, float z, float w)
	{ vertex(Vector4(x, y, z, w)); }

	void vertex(const Vector3 &v)
	{ vertex(Vector4(v.x, v.y, v.z, 1)); }

	void vertex(const Vector4 &v)
	{ vertex_(mtx*v); }

protected:
	virtual void vertex_(const Vector4 &) = 0;

public:
	void attrib(unsigned i, const Vector4 &v)
	{
		if(i>=attr.size())
			attr.resize(i+1);
		attr[i] = v;
	}

	void normal(float x, float y, float z)
	{ normal(Vector3(x, y, z)); }

	void normal(const Vector3 &n)
	{ attrib(get_attribute_semantic(NORMAL3), mtx*Vector4(n.x, n.y, n.z, 0)); }

	void tangent(float x, float y, float z)
	{ tangent(Vector3(x, y, z)); }

	void tangent(const Vector3 &t)
	{ attrib(get_attribute_semantic(TANGENT3), mtx*Vector4(t.x, t.y, t.z, 0)); }

	void binormal(float x, float y, float z)
	{ binormal(Vector3(x, y, z)); }

	void binormal(const Vector3 &b)
	{ attrib(get_attribute_semantic(BINORMAL3), mtx*Vector4(b.x, b.y, b.z, 0)); }

	void texcoord(float s)
	{ texcoord(s, 0, 0, 1); }

	void texcoord(float s, float t)
	{ texcoord(s, t, 0, 1); }

	void texcoord(float s, float t, float r)
	{ texcoord(s, t, r, 1); }

	void texcoord(float s, float t, float r, float q)
	{ texcoord(Vector4(s, t, r, q)); }

	void texcoord(const Vector4 &t)
	{ multitexcoord(0, t); }

	void multitexcoord(unsigned i, float s)
	{ multitexcoord(i, s, 0, 0, 1); }

	void multitexcoord(unsigned i, float s, float t)
	{ multitexcoord(i, s, t, 0, 1); }

	void multitexcoord(unsigned i, float s, float t, float r)
	{ multitexcoord(i, s, t, r, 1); }

	void multitexcoord(unsigned i, float s, float t, float r, float q)
	{ multitexcoord(i, Vector4(s, t, r, q)); }

	void multitexcoord(unsigned i, const Vector4 &t)
	{ attrib(get_attribute_semantic(TEXCOORD4)+i, t); }

	void color(unsigned char r, unsigned char g, unsigned char b)
	{ color(r, g, b, 255); }

	void color(unsigned char r, unsigned char g, unsigned char b, unsigned char a)
	{ color(r/255.f, g/255.f, b/255.f, a/255.f); }

	void color(float r, float g, float b)
	{ color(r, g, b, 1); }

	void color(float r, float g, float b, float a)
	{ color(Color(r, g, b, a)); }

	void color(const Color &c)
	{ attrib(get_attribute_semantic(COLOR4_FLOAT), Vector4(c.r, c.g, c.b, c.a)); }

	void group(int g0)
	{ group(g0, 0, 0, 0); }

	void group(int g0, int g1)
	{ group(g0, g1, 0, 0); }

	void group(int g0, int g1, int g2)
	{ group(g0, g1, g2, 0); }

	void group(int g0, int g1, int g2, int g3)
	{ attrib(get_attribute_semantic(GROUP4), Vector4(g0, g1, g2, g3)); }

	void weight(float w0)
	{ weight(w0, 0, 0, 0); }

	void weight(float w0, float w1)
	{ weight(w0, w1, 0, 0); }

	void weight(float w0, float w1, float w2)
	{ weight(w0, w1, w2, 0); }

	void weight(float w0, float w1, float w2, float w3)
	{ attrib(get_attribute_semantic(WEIGHT4), Vector4(w0, w1, w2, w3)); }

	void generic(unsigned i, float x)
	{ generic(i, x, 0, 0, 1); }

	void generic(unsigned i, float x, float y)
	{ generic(i, x, y, 0, 1); }

	void generic(unsigned i, float x, float y, float z)
	{ generic(i, x, y, z, 1); }

	void generic(unsigned i, float x, float y, float z, float w)
	{ generic(i, Vector4(x, y, z, w)); }

	void generic(unsigned i, const Vector4 &a)
	{ attrib(get_attribute_semantic(GENERIC4)+i, a); }
};

} // namespace GL
} // namespace Msp

#endif
